# Copyright Contributors to the Opensynth-energy Project.
# SPDX-License-Identifier: Apache-2.0

from typing import Optional, Tuple, TypedDict

import pytorch_lightning as pl
import torch
import torch.nn as nn
from pytorch_lightning.callbacks import EarlyStopping

from opensynth.models.faraday.gaussian_mixture import gmm_metrics, gmm_utils
from opensynth.models.faraday.vae_model import FaradayVAE


class GMMInitParams(TypedDict):
    labels: torch.Tensor
    means: torch.Tensor
    responsibilities: torch.Tensor
    weights: torch.Tensor
    covariances: torch.Tensor
    precision_cholesky: torch.Tensor


class GaussianMixtureModel(nn.Module):

    weights: torch.Tensor
    means: torch.Tensor
    precision_cholesky: torch.Tensor
    covariances: torch.Tensor
    nll: torch.Tensor

    def __init__(
        self,
        num_components: int,
        num_features: int,
        reg_covar: float = 1e-6,
    ):

        super().__init__()
        self.num_components = num_components
        self.num_features = num_features
        self.reg_covar = reg_covar

        # Initialise model params
        weights_shape = torch.Size([self.num_components])
        means_shape = torch.Size([self.num_components, self.num_features])
        precision_cholesky_shape = torch.Size(
            [self.num_components, self.num_features, self.num_features]
        )
        covariances_shape = torch.Size(
            [self.num_components, self.num_features, self.num_features]
        )
        nll_shape = torch.Size([1])
        self.register_buffer("weights", torch.empty(weights_shape))
        self.register_buffer("means", torch.empty(means_shape))
        self.register_buffer(
            "precision_cholesky", torch.empty(precision_cholesky_shape)
        )
        self.register_buffer("covariances", torch.empty(covariances_shape))
        self.register_buffer("nll", torch.empty(nll_shape))
        self.initialised = False

    def initialise(self, init_params: GMMInitParams):
        self.means.data = init_params["means"]
        self.precision_cholesky.data = init_params["precision_cholesky"]
        self.weights.data = init_params["weights"]
        self.initialised = True

    @staticmethod
    def _compute_log_det_cholesky(
        matrix_chol: torch.Tensor, n_features: int
    ) -> torch.Tensor:
        """
        Compute the log-det of the cholesky decomposition of matrices.
        Pytorch implementation of sklearn's
        sklearn.mixture._gaussian_mixture._compute_log_det_cholesky

        Args:
            matrix_chol (torch.Tensor): Cholesky matrix
            n_features (int): Number of features

        Returns:
            torch.Tensor: Log determinant of cholesky matrix
        """
        n_components, _, _ = matrix_chol.shape
        log_det_chol = torch.sum(
            torch.log(
                matrix_chol.reshape(n_components, -1)[:, :: n_features + 1]
            ),
            dim=1,
        )
        log_det_chol = log_det_chol.to(matrix_chol.device)
        return log_det_chol

    def _estimate_log_gaussian_prob(
        self,
        X: torch.Tensor,
    ) -> torch.Tensor:
        """
        Estimate the log gaussian probability.
        Pytorch implementation of sklearn's
        sklearn.mixture._gaussian_mixture._estimate_log_gaussian_prob

        Args:
            X (torch.Tensor): Input data

        Returns:
            torch.Tensor: Log probabability
        """
        if self.initialised is False:
            raise AttributeError("Model is not initialised.")
        n_samples, n_features = X.shape
        # Log determinant of cholesky matrix
        log_det = self._compute_log_det_cholesky(
            self.precision_cholesky, n_features
        )
        # Log of probabilities
        log_prob = torch.empty(
            (n_samples, self.num_components), device=X.device
        )
        for k, (mu, prec_chol) in enumerate(
            zip(self.means, self.precision_cholesky)
        ):
            y = torch.matmul(X, prec_chol) - torch.matmul(mu, prec_chol)
            log_prob[:, k] = torch.sum(torch.square(y), dim=1)
        # log gaussian likelihood

        pi = torch.tensor(torch.pi, device=log_prob.device)
        return -0.5 * (n_features * torch.log(2 * pi) + log_prob) + log_det

    def _estimate_log_weights(self: torch.Tensor) -> torch.Tensor:
        """
        Estimate log of weights.
        Pytorch implementation of sklearns's
        sklearn.mixture._base.BaseMixture._estimate_log_weights

        Returns:
            torch.Tensor: Log of weights
        """
        if self.initialised is False:
            raise AttributeError("Model is not initialised.")

        return torch.log(self.weights)

    def _estimate_weighted_log_prob(
        self,
        X: torch.Tensor,
    ) -> torch.Tensor:
        """
        Estimated weighted log probability.
        Pytorch's implementation of sklearn's
        sklearn.mixture._base.BaseMixture._estimate_weighted_log_prob

        Args:
            X (torch.Tensor): Input data

        Returns:
            torch.Tensor: Weighted log probability
        """
        w = self._estimate_log_weights()
        p = self._estimate_log_gaussian_prob(X)
        return p + w

    def _estimate_log_prob_and_responsibilities(
        self,
        X: torch.tensor,
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        """
        Compute the log probabilities, weighted log probabilities per
        component and responsibilities for each sample in X with respect to
        the current state of the model.

        Pytorch implementation of sklearn's
        sklearn.mixture._base.BaseMixture._estimate_log_prob_resp
        Args:
            X (torch.tensor): Input data

        Returns:
            Tuple[torch.Tensor, torch.Tensor]:
                Normalised log probabilities and log responsibilities
        """
        weighted_log_prob = self._estimate_weighted_log_prob(X)
        log_prob_norm = torch.logsumexp(weighted_log_prob, dim=1)
        log_resp = weighted_log_prob - log_prob_norm.reshape(-1, 1)
        return log_prob_norm, log_resp

    def e_step(
        self,
        X,
    ):
        log_prob_norm, log_resp = self._estimate_log_prob_and_responsibilities(
            X
        )
        return torch.mean(log_prob_norm), log_resp

    def m_step(self, X: torch.Tensor, log_reponsibilities: torch.Tensor):
        weights_, means_, covariances_ = (
            gmm_utils.torch_estimate_gaussian_parameters(
                X,
                responsibilities=torch.exp(log_reponsibilities),
                reg_covar=self.reg_covar,
            )
        )
        precision_cholesky_ = gmm_utils.torch_compute_precision_cholesky(
            covariances=covariances_, reg=self.reg_covar
        )
        return precision_cholesky_, weights_, means_, covariances_

    def update_params(
        self,
        weights: torch.Tensor,
        means: torch.Tensor,
        precision_cholesky: torch.Tensor,
        covariances: torch.Tensor,
        nll: torch.Tensor,
    ):
        self.weights.data = weights
        self.means.data = means
        self.precision_cholesky.data = precision_cholesky
        self.covariances.data = covariances
        self.nll.data = nll
        return self

    def forward(self, X: torch.Tensor):
        return self.e_step(X)

    def predict(self, X: torch.Tensor):
        return self._estimate_weighted_log_prob(X).argmax(dim=1)

    def sample(self, n_samples: int) -> torch.Tensor:
        """Sample from GMM components

        Args:
            n_samples (int): number of samples to generate

        Returns:
            torch.Tensor: samples drawn from GMM size n_samples x n_components
        """

        # Set up the random generator
        generator = torch.Generator()
        # Sample component counts from the multinomial distribution
        n_samples_comp = torch.multinomial(
            self.weights, n_samples, replacement=True, generator=generator
        ).bincount(minlength=len(self.weights))

        # Initialize lists to collect samples and labels
        X = []
        y = []

        # Sample from each component based on the number of samples
        for j, (mean, covariance, sample_count) in enumerate(
            zip(self.means, self.covariances, n_samples_comp)
        ):
            if (
                sample_count > 0
            ):  # Only sample if we need samples from this component
                dist = torch.distributions.MultivariateNormal(mean, covariance)
                samples = dist.sample((sample_count,))
                X.append(samples)
                y.append(torch.full((sample_count,), j, dtype=torch.int64))

        # Concatenate all samples and labels into single tensors
        X = torch.vstack(X)
        y = torch.cat(y)

        return X


class GaussianMixtureLightningModule(pl.LightningModule):

    def __init__(
        self,
        gmm_module: GaussianMixtureModel,
        vae_module: FaradayVAE,
        num_components: int,
        num_features: int,
        reg_covar: float = 1e-6,
        convergence_tolerance: float = 1e-2,
        sync_on_batch: bool = False,
        sample_weights_column: Optional[str] = None,
    ):
        super().__init__()
        self.gmm_module = gmm_module
        self.vae_module = vae_module
        self.num_components = num_components
        self.num_features = num_features
        self.reg_covar = reg_covar

        self.automatic_optimization = False
        self.convergence_tolerance = convergence_tolerance

        # GMM params to sync across processes
        self.weight_metric = gmm_metrics.WeightsMetric(self.num_components)
        self.mean_metric = gmm_metrics.MeansMetric(
            self.num_components, self.num_features
        )
        self.precision_cholesky_metric = gmm_metrics.PrecisionCholeskyMetric(
            self.num_components, self.num_features
        )
        self.covariance_metric = gmm_metrics.CovarianceMetric(
            self.num_components, self.num_features
        )
        self.nll = gmm_metrics.NegativeLogLikelihoodMetric()

        self.sync_on_batch = sync_on_batch

        self.sample_weights_column = sample_weights_column

    def configure_optimizers(self) -> None:
        return None

    def on_train_epoch_start(self) -> None:
        # At the start of epoch, reset metrics
        self.mean_metric.reset()
        self.weight_metric.reset()
        self.precision_cholesky_metric.reset()
        self.covariance_metric.reset()
        self.nll.reset()

    def training_step(self, batch) -> None:
        # Encode the batch
        encoded_batch = gmm_utils.prepare_data_for_training_step(
            batch, self.vae_module, self.sample_weights_column
        )

        # Run e-step
        log_prob, log_resp = self.gmm_module.e_step(encoded_batch)
        # Run m-step
        precision_cholesky, weights, means, covariances = (
            self.gmm_module.m_step(encoded_batch, log_resp)
        )
        # Update model params. This only updates the params
        # on the current device
        self.gmm_module.update_params(
            weights=weights,
            means=means,
            precision_cholesky=precision_cholesky,
            covariances=covariances,
            nll=torch.neg(log_prob),
        )

    def on_train_batch_end(self, *args, **kwargs) -> None:
        """If `sync_on_batch` is True, sync model parameters across devices at
        the end of each batch. Otherwise, sync is done at the end of epoch.
        """

        if self.sync_on_batch:
            weights = self.gmm_module.weights
            means = self.gmm_module.means
            precision_cholesky = self.gmm_module.precision_cholesky
            covariances = self.gmm_module.covariances
            nll = self.gmm_module.nll

            # forward performs update, compute and reset metrics
            weights_reduced = self.weight_metric.forward(weights)
            means_reduced = self.mean_metric.forward(means)
            prec_chol_reduced = self.precision_cholesky_metric.forward(
                precision_cholesky
            )
            covar_reduced = self.covariance_metric.forward(covariances)
            nll_reduced = self.nll.forward(nll)

            self.gmm_module.update_params(
                weights=weights_reduced,
                means=means_reduced,
                precision_cholesky=prec_chol_reduced,
                covariances=covar_reduced,
                nll=nll_reduced,
            )
            self.log(
                "nll",
                nll_reduced,
                on_step=False,
                on_epoch=True,
            )  # uses mean-reduction (default) to accumulate the metrics

    def on_train_epoch_end(self) -> None:
        # At the end of epoch, update metrics and sync across
        # multiple devices using torchmetrics.Metric.compute
        # Then update model params using the synced values

        if not self.sync_on_batch:
            weights = self.gmm_module.weights
            means = self.gmm_module.means
            precision_cholesky = self.gmm_module.precision_cholesky
            covariances = self.gmm_module.covariances
            nll = self.gmm_module.nll

            self.weight_metric.update(weights)
            self.mean_metric.update(means)
            self.precision_cholesky_metric.update(precision_cholesky)
            self.covariance_metric.update(covariances)
            self.nll.update(nll)

            weights_reduced = self.weight_metric.compute()
            means_reduced = self.mean_metric.compute()
            prec_chol_reduced = self.precision_cholesky_metric.compute()
            covar_reduced = self.covariance_metric.compute()
            nll_reduced = self.nll.compute()

            self.log(
                "nll",
                nll_reduced,
                on_step=False,
                on_epoch=True,
            )  # uses mean-reduction (default) to accumulate the metrics

            self.gmm_module.update_params(
                weights=weights_reduced,
                means=means_reduced,
                precision_cholesky=prec_chol_reduced,
                covariances=covar_reduced,
                nll=nll_reduced,
            )

    def configure_callbacks(self) -> list[pl.Callback]:
        early_stopping = EarlyStopping(
            "nll",
            min_delta=self.convergence_tolerance,
            patience=1,
            # check_on_train_epoch_end=True,
            mode="min",
        )
        return [early_stopping]



# Copyright Contributors to the Opensynth-energy Project.
# SPDX-License-Identifier: Apache-2.0

import logging
import sys
from dataclasses import dataclass

import matplotlib.pyplot as plt
import pandas as pd
import seaborn as sns
import torch

from opensynth.data_modules.lcl_data_module import LCLDataModule
from opensynth.evaluation.privacy import generate_attack_data
from opensynth.models.faraday import FaradayModel

logging.basicConfig(stream=sys.stdout, level=logging.INFO)
logger = logging.getLogger(__name__)


@dataclass
class ReconstructionAttackDataset:
    synthetic_samples: torch.Tensor
    real_samples: torch.Tensor
    outlier_samples: torch.Tensor


def create_attack_dataset(
    model: FaradayModel,
    dm: LCLDataModule,
    n_samples: int = 20000,
) -> ReconstructionAttackDataset:
    """
    Generate reconstruction attack dataset

    Args:
        model (FaradayModel): Synthetic model
        dm (LCLDataModule): LCL data module
        n_samples (int): Number of samples to generate. Defaults to 20K.

    Returns:
        ReconstructionAttackDataset: Reconstruction attack data.
    """
    synthetic_samples = generate_attack_data.generate_synthetic_samples(
        model, dm, n_samples
    )

    real_samples = generate_attack_data.draw_real_data(
        dm, n_samples, outliers=False
    )

    outlier_samples = generate_attack_data.draw_real_data(
        dm, n_samples, outliers=True
    )

    return ReconstructionAttackDataset(
        synthetic_samples=synthetic_samples,
        real_samples=real_samples,
        outlier_samples=outlier_samples,
    )


def _calculate_pairwise_euclidean(
    vec1: torch.Tensor, vec2: torch.Tensor
) -> torch.Tensor:
    """
        Calculates pairwise euclidean distance. Creates a matrix of
        Size(Vec1) X Size(Vec2)
    Args:
        vec1 (torch.Tensor): Vector 1
        vec2 (torch.Tensor): Vector 2

    Returns:
        torch.Tensor: Pairwise euclidean distance matrix
    """
    logger.info("Calculating pairwise euclidean distance..")
    euc_pairwise = torch.cdist(vec1, vec2)
    return euc_pairwise


def _convert_wide_to_long(
    matrix: torch.Tensor, metric_name: str
) -> pd.DataFrame:
    """
    Converts euclidean matrix (wide table) to long table of
    pairwise distances

    Args:
        matrix (torch.Tensor): Pairwise euclidean distance matrix
        metric_name (str): Metric name

    Returns:
        pd.DataFrame: A long dataframe consisting of the seen outlier
        and its pairwise distances to all generated data.
    """
    logger.info("Converting wide to long table..")
    df = pd.DataFrame(matrix.detach().numpy())
    df.columns = [f"{metric_name}{i + 1}" for i in range(matrix.shape[1])]
    df["real_outlier"] = df.index + 1

    df_out = pd.wide_to_long(
        df, stubnames=metric_name, i=["real_outlier"], j="synthetic_output"
    )
    df_out = df_out.reset_index()
    df_out = df_out.sort_values(
        by=["real_outlier", "synthetic_output"], ascending=True
    )
    return df_out


def _calculate_vector_norm(data: torch.Tensor) -> pd.DataFrame:
    """
    Calculate norm of outlier vector

    Args:
        data (torch.Tensor): Input data

    Returns:
        pd.DataFrame: A long dataframe consisting of the seen outlier
        and its pairwise distances to all generated data.
    """
    logger.info("Calculating vector norm..")
    df_vector_norm = pd.DataFrame(columns=["real_outlier"])
    df_vector_norm["real_outlier"] = [i + 1 for i in range(len(data))]
    df_vector_norm["norm"] = torch.linalg.vector_norm(data, dim=1)
    return df_vector_norm


def calculate_distance_norm(
    real: torch.Tensor, fake: torch.Tensor, group_min: bool = True
) -> pd.DataFrame:
    """
    Calculates distance norms between generated data and seen outliers.

    Args:
        real (torch.Tensor): Seen outlier
        fake (torch.Tensor): Generated data
        group_min (bool, optional): Group pairwise distances to
        return only the nearest neighbour. Defaults to True. If
        False, returns all pairwise distances.

    Returns:
        pd.DataFrame: Pandas dataframe of outliers to its
        neighbour. Distance is expressed as a ratio of
        the outlier's vector norm.
    """
    matrix = _calculate_pairwise_euclidean(real, fake)
    df_euc = _convert_wide_to_long(matrix, metric_name="euc_dist")
    df_norm = _calculate_vector_norm(real)

    df_dist = df_euc.merge(df_norm, on="real_outlier", how="left")
    df_dist["ratio"] = df_dist["euc_dist"] / df_dist["norm"]
    df_dist = df_dist.sort_values(by="ratio", ascending=True)

    if group_min:
        df_dist = (
            df_dist.groupby(["real_outlier"])["ratio"].min().reset_index()
        )

    return df_dist


def plot_ecdf(df_ecdf, hue_col: str | None = None):
    """
    Plots the cumulative distritution function for reconstruction
    attack.

    Args:
        df_ecdf (_type_): Dataframe
        hue_col (str, optional): Column to groupby. Defaults to None.
    """
    sns.ecdfplot(data=df_ecdf, x="ratio", hue=hue_col)
    plt.xlabel("Ratio of Euclidean Distance to Vector Norm")
    plt.ylabel("Proportion")
    plt.xlim(0, 1)



# Copyright Contributors to the Opensynth-energy Project.
# SPDX-License-Identifier: Apache-2.0

import logging
from pathlib import Path
from typing import Optional

from opensynth.datasets import datasets_utils
from opensynth.datasets.low_carbon_london import (
    preprocess_lcl,
    split_households,
)

logger = logging.getLogger(__name__)


def download_lcl_data(data_dir: str = "./data"):
    """
    Download the Low Carbon London (LCL) dataset.
    The LCL dataset is a large dataset and may take a while to download.
    The compressed version of LCL dataset is about 700MB, but the full
    dataset size is about 8GB.
    Decompression may not work on windows. If so, please manually download
    and unzip the contents into the `{data_dir}/raw` folder.
    Args:
        data_dir (str): Directory to save the downloaded data. Defaults to
            "./data".
    Returns:
        None
    """
    LCL_URL = "https://data.london.gov.uk/download/smartmeter-energy-use-data-in-london-households/3527bf39-d93e-4071-8451-df2ade1ea4f2/LCL-FullData.zip"  # noqa
    FILE_NAME = Path(f"{data_dir}/raw/lcl_full_data.zip")  # noqa
    datasets_utils.download_data(LCL_URL, FILE_NAME)


def split_preprocess_data(
    data_dir: str,
    csv_data_path: str,
    sample_fraction: float,
    time_resolution: str,
    feature_cols: list[str],
    id_col: str,
    kwh_col: str,
    datetime_col: str,
    utc: bool,
    datetime_format: Optional[str],
    historical_start: str,
    historical_end: str,
    future_start: str,
    future_end: str,
    drop_nulls: bool,
):
    """
    Read in, split and preprocess the dataset.

    Requirements:
    - Dataset needs to be stored in CSV format in {data_dir}/{csv_data_path}.
    - Dataset needs to have an ID column, a kWh column and a datetime column.
    - If time_resolution is "half_hourly", data needs to be at half-hourly
      resolution. If "hourly", data needs to be at hourly resolution.
    - Datetime column needs to be in a format that can be parsed by
        pandas.to_datetime, if datetime_format is specified, then this will be
        parsed into pandas.to_datetime.
    - kWh column needs to be in numeric format.
    - feature_cols needs to be a list of columns in the dataset to include as
      features. These can be categorical or numeric.
    - historical_start, historical_end, future_start and future_end define
        the historical and future periods for for TSTR evaluation. Needs to be
        in the format "YYYY-MM-DD".
    Args:
        data_dir (str): Directory to save the processed data.
        csv_data_path (str): Path to the CSV file containing the dataset.
        sample_fraction (float): Fraction of households to include in the
            training set.
        time_resolution (str): Time resolution of the data. Allowed values:
            "half_hourly", "hourly".
        feature_cols (List[str]): List of feature columns to include.
        id_col (str): Name of the household ID column.
        kwh_col (str): Name of the kWh column.
        datetime_col (str): Name of the datetime column.
        utc (bool): Whether to parse datetime as UTC.
        datetime_format (str, optional): Format of the datetime column.
        historical_start (str): Start date for historical data.
        historical_end (str): End date for historical data.
        future_start (str): Start date for future data.
        future_end (str): End date for future data.
        drop_nulls (bool): Whether to drop rows with NaN kwh values. If False,
            will replace NaN kwh values with 0.0
    """

    CSV_FILE_NAME = Path(f"{data_dir}/{csv_data_path}")
    logger.info(
        f"Reading data from {CSV_FILE_NAME}. Storing data in {data_dir}."
    )

    # Split dataset into training/ holdout sets
    split_households.split_data(
        data_dir,
        CSV_FILE_NAME,
        sample_fraction=sample_fraction,
        id_col=id_col,
        kwh_col=kwh_col,
        datetime_col=datetime_col,
        utc=utc,
        datetime_format=datetime_format,
        historical_start=historical_start,
        historical_end=historical_end,
        future_start=future_start,
        future_end=future_end,
    )
    # Preprocess the data into daily load profiles
    preprocess_lcl.preprocess_data(
        data_dir,
        datetime_col=datetime_col,
        kwh_col=kwh_col,
        id_col=id_col,
        utc=utc,
        datetime_format=datetime_format,
        time_resolution=time_resolution,
        feature_cols=feature_cols,
        drop_nulls=drop_nulls,
    )


if __name__ == "__main__":
    data_dir = "./data"

    sample_fraction = 0.75

    # Dataset location
    csv_data_path = "raw/CC_LCL-FullData.csv"

    # Dataset parameters
    # Adjust these parameters if using a different dataset
    # Parameters defined here are for the LCL dataset
    time_resolution = "half_hourly"
    feature_cols = ["stdorToU"]
    id_col = "LCLid"
    kwh_col = "KWH/hh (per half hour) "
    date_col = "DateTime"
    utc = False
    datetime_format = None
    historical_start = "2012-01-01"
    historical_end = "2013-12-31"
    future_start = "2014-01-01"
    future_end = "2014-12-31"
    drop_nulls = True

    split_preprocess_data(
        data_dir,
        csv_data_path,
        sample_fraction,
        time_resolution,
        feature_cols,
        id_col,
        kwh_col,
        date_col,
        utc,
        datetime_format,
        historical_start,
        historical_end,
        future_start,
        future_end,
        drop_nulls,
    )



# Copyright Contributors to the Opensynth-energy Project.
# SPDX-License-Identifier: Apache-2.0

import logging
import os
import random
from pathlib import Path
from typing import List, Optional, Tuple

import numpy as np
import pandas as pd

random.seed(42)

logger = logging.getLogger(__name__)


def split_household_ids(
    df: pd.DataFrame,
    id_col: str,
    sample_fraction: float = 0.75,
) -> Tuple[List[str], List[str]]:
    """
    Split dataset into training vs holdout households.

    Args:
        df (pd.DataFrame): dataset
        id_col (str): Name of the household ID column
        sample_fraction (float): Fraction of household ids to include in
        training set

    Returns:
        Tuple[List[str], List[str]]: List of training and holdout household ids
    """
    logger.info("Splitting households into train and holdout households")
    unique_ids = df[id_col].unique().tolist()
    random.shuffle(unique_ids)
    sample_size = int(len(unique_ids) * sample_fraction)

    train_ids = unique_ids[:sample_size]
    holdout_ids = unique_ids[sample_size:]

    return train_ids, holdout_ids


def split_historical_future_periods(
    df: pd.DataFrame,
    datetime_col: str,
    historical_start: str,
    historical_end: str,
    future_start: str,
    future_end: str,
) -> Tuple[pd.DataFrame, pd.DataFrame]:
    """
    Splits dataset into:
     1) historical period
     2) future period
    These periods are used for TSTR evaluation.

    Args:
        df (pd.DataFrame): pd.DataFrame
        datetime_col (str): Name of the datetime column
        historical_start (str): Start date for historical period
        historical_end (str): End date for historical period
        future_start (str): Start date for future period
        future_end (str): End date for future period

    Returns:
        Tuple[pd.DataFrame, pd.DataFrame]: Historical and Future dataframe.
    """
    # Format the start and end dates to include time information
    historical_start = pd.Timestamp(historical_start).replace(
        hour=00, minute=00, second=00
    )
    historical_end = pd.Timestamp(historical_end).replace(
        hour=23, minute=59, second=59
    )
    future_start = pd.Timestamp(future_start).replace(
        hour=00, minute=00, second=00
    )
    future_end = pd.Timestamp(future_end).replace(
        hour=23, minute=59, second=59
    )

    historical_start_mask = df[datetime_col] >= historical_start
    historical_end_mask = df[datetime_col] <= historical_end
    tstr_historical_mask = historical_start_mask & historical_end_mask

    future_start_mask = df[datetime_col] >= future_start
    future_end_mask = df[datetime_col] <= future_end
    tstr_future_mask = future_start_mask & future_end_mask

    df_historical = df.loc[tstr_historical_mask]
    df_future = df.loc[tstr_future_mask]
    return df_historical, df_future


def split_data(
    data_dir: str,
    csv_filename: Path,
    sample_fraction: float = 0.75,
    id_col: str = "ID",
    kwh_col: str = "kwh",
    datetime_col: str = "DateTime",
    utc=True,
    datetime_format: Optional[str] = None,
    historical_start: str = "2012-01-01",
    historical_end: str = "2013-12-31",
    future_start: str = "2014-01-01",
    future_end: str = "2015-12-31",
) -> None:
    """
    Split the dataset 4 ways:
    1) Historical Train household data
    2) Historical Holdout household data
    3) Future Train household data
    4) Future Holdout household data

    Historical data is used for training generative models.
    Future data is used for Train-Synthetic-Test-Real (TSTR) evaluation.

    Args:
        data_dir (str): Directory to store the processed data.
        csv_filename (Path): Path to the CSV file containing the raw data.
        sample_fraction (float): Fraction of households to include in the
            training set. Defaults to 0.75.
        id_col (str): Name of the household ID column. Defaults to "LCLid".
        kwh_col (str): Name of the kWh column. Defaults to
            "KWH/hh (per half hour) ".
        datetime_col (str): Name of the datetime column. Defaults to
            "DateTime".
        utc (bool): Whether to parse datetime as UTC. Defaults to False.
        datetime_format (str, optional): Format of the datetime column.
            Defaults to None.
        historical_start (str): Start date for historical data. Defaults to
            "2012-01-01".
        historical_end (str): End date for historical data. Defaults to
            "2013-12-31".
        future_start (str): Start date for future data. Defaults to
            "2014-01-01".
        future_end (str): End date for future data. Defaults to "2014-12-31".
    Returns:
        None
    """

    logger.info(f"👀 Reading data from: {csv_filename}")
    df = pd.read_csv(csv_filename)

    logger.info("🧹 Formatting data")
    df[datetime_col] = pd.to_datetime(
        df[datetime_col], utc=utc, format=datetime_format
    )
    df[kwh_col] = df[kwh_col].replace("Null", np.float64())
    df[kwh_col] = df[kwh_col].astype(float)
    df[id_col] = df[id_col].astype(str)

    logger.info("🖖 Spliting households into train and holdout")
    train_ids, holdout_ids = split_household_ids(
        df,
        id_col,
        sample_fraction=sample_fraction,
    )
    logger.info(f"Train len: {len(train_ids)}")
    logger.info(f"Holdout len: {len(holdout_ids)}")

    logger.info("📆 Splitting data into train and holdout period")
    df_history, df_future = split_historical_future_periods(
        df,
        datetime_col,
        historical_start,
        historical_end,
        future_start,
        future_end,
    )
    logger.info(f"History len: {len(df_history)}")
    logger.info(f"Future len: {len(df_future)}")

    df_historical_train = df_history[df_history[id_col].isin(train_ids)]
    df_future_train = df_future[df_future[id_col].isin(train_ids)]
    df_historical_holdout = df_history[df_history[id_col].isin(holdout_ids)]
    df_future_holdout = df_future[df_future[id_col].isin(holdout_ids)]

    logger.info("📦 Saving train and holdout data")
    historical_path = Path(f"{data_dir}/raw/historical")
    future_path = Path(f"{data_dir}/raw/future")
    os.makedirs(historical_path, exist_ok=True)
    os.makedirs(future_path, exist_ok=True)

    df_historical_train.to_csv(
        f"{historical_path}/train.csv",
        index=False,
    )
    df_historical_holdout.to_csv(
        f"{historical_path}/holdout.csv",
        index=False,
    )
    df_future_train.to_csv(f"{future_path}/train.csv", index=False)
    df_future_holdout.to_csv(f"{future_path}/holdout.csv", index=False)

    logger.info("👍 Done!")



import os
import pathlib

from setuptools import find_packages, setup

# allow setup.py to be run from any path
os.chdir(os.path.normpath(os.path.join(os.path.abspath(__file__), os.pardir)))


VERSION = "v1.0.0"

REPO_ROOT = pathlib.Path(__file__).parent

# Fetch the long description from the readme
with open(REPO_ROOT / "README.md", encoding="utf-8") as f:
    README = f.read()

install_requires = (
    "wget>=3.2",
    "tqdm>=4.66.4",
    "numpy<=2.0.0",
    "torch>=2.3.1",
    "scikit-learn>=1.3.2",
    "pytorch-lightning>=2.3.3",
    "matplotlib>3.6.1",
    "opacus>=1.4.1",
    "seaborn>=0.13.2",
    "torchmetrics>=1.4.1",
    "pandas>=1.2",
    "litdata==0.2.30",
    "zstd==1.5.5.1",
    "einops>=0.7.0",
    "ema-pytorch==0.4.5",
)

setup(
    name="opensynth-energy",
    version=VERSION,
    include_package_data=True,
    description="Opensynth is a library for"
    "synthetic energy demand generation.",
    long_description=README,
    long_description_content_type="text/markdown",
    author="Opensynth",
    author_email="pypi@opensynth.energy",
    license="Apache 2.0",
    package_dir={"": "src"},
    packages=find_packages("src"),
    install_requires=install_requires,
    process_dependency_links=True,
    classifiers=[
        "Intended Audience :: Developers",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.11",
    ],
)



# Copyright Contributors to the Opensynth-energy Project.
# SPDX-License-Identifier: Apache-2.0

import csv
import logging
import os
from pathlib import Path
from typing import List, Optional, Tuple

import numpy as np
import pandas as pd

from opensynth.datasets.datasets_utils import NoiseFactory, NoiseType

logger = logging.getLogger(__name__)


def get_current_month_end(df: pd.DataFrame, date_col="dt"):
    """
    Get the month end of current month.
    Note: Pandas offsets.MonthEnd()
    returns the date of the following month instead of the current month.

    E.g. if dt = 31st Jan, offsets.MonthEnd returns 28th Feb
    instead of 31st Jan. To work around this, we offset by -1 days
    to get previous day, then do offsets.MonthEnd()

    This is faster than offsets.MonthEnd().rollforward(ts):
    - This method doesn't work on array, only works on raw Timestamp object,
    which will require for loops (or .apply) which is super slow on big data
    - More efficient to perform this in a vectorised manner by offset
    with -1 days then use offsets.MonthEnd().

    Args:
        df (Pandas dataframe): Input pandas dataframe
        date_col (str, optional): Name of date column. Defaults to "dt".

    Returns:
        pd.DataFrame: Output dataframe with the "month_end" column
    """
    df["month_end"] = df[date_col] + pd.offsets.Day(-1)
    df["month_end"] = df["month_end"] + pd.offsets.MonthEnd()
    df["month_end"] = df["month_end"].dt.date
    return df


def load_data(file_path: Path) -> pd.DataFrame:
    """
    Load data from csv

    Args:
        file_path (Path): Path to the csv file

    Returns:
        pd.DataFrame: dataset
    """
    logger.info(f"🚛 Loading data from {file_path}")
    df = pd.read_csv(file_path)
    return df


def format_data(
    df: pd.DataFrame,
    datetime_col: str,
    kwh_col: str,
    id_col: str,
    utc: bool = True,
    datetime_format: Optional[str] = None,
) -> pd.DataFrame:
    """
    Load data from csv

    Args:
        df (pd.DataFrame): Dataset
        datetime_col (str): Name of the datetime column
        kwh_col (str): Name of the kWh column
        id_col (str): Name of the household ID column
        utc (bool): Whether the datetime is in UTC
        datetime_format (str, optional): Format of the datetime column. If
            None, will try to infer the format. Defaults to None.

    Returns:
        pd.DataFrame: dataset
    """

    logger.info("🧹 Formatting data")
    df.rename(
        columns={datetime_col: "DateTime", kwh_col: "kwh", id_col: "ID"},
        inplace=True,
    )
    df["DateTime"] = pd.to_datetime(
        df["DateTime"], utc=utc, format=datetime_format
    )
    df["kwh"] = df["kwh"].replace("Null", np.nan)  # Replace "Null" with np.nan
    df["kwh"] = df["kwh"].astype(float)
    df["ID"] = df["ID"].astype(str)
    return df


def extract_date_features(
    df: pd.DataFrame,
) -> pd.DataFrame:
    """
    Parse date features from DateTime column.

    Args:
        df (pd.DataFrame): Input dataframe

    Returns:
        pd.DataFrame: Output dataframe with date features
    """
    logger.info("📅 Extracting date features")
    df["dt"] = pd.to_datetime(df["DateTime"])
    df["date"] = df["dt"].dt.date.astype(str)
    df["month"] = df["dt"].dt.month.astype(int)
    df["week"] = df["dt"].dt.to_period("W-MON").dt.start_time
    df = get_current_month_end(df, date_col="dt")
    df["month_max"] = pd.to_datetime(df["month_end"]).dt.day.astype(int)
    df["day"] = df["dt"].dt.day.astype(int)
    df["dayofweek"] = df["dt"].dt.dayofweek.astype(int)
    df["hour"] = df["dt"].dt.hour.astype(int)
    df["minute"] = df["dt"].dt.minute.astype(int)
    return df


def parse_settlement_period(
    df: pd.DataFrame,
) -> pd.DataFrame:
    """
    Parse settlement periods from hour and minute columns

    Args:
        df (pd.DataFrame): Input dataframe

    Returns:
        pd.DataFrame: Output dataframe with settlement period column
    """
    logger.info("🕰 Parsing Settlement Period")

    def _get_settlement_offset(minute_value):
        if minute_value >= 30:
            return 1
        return 0

    df_out = df.copy()
    df_out["settlement_offset"] = df_out["minute"].apply(
        _get_settlement_offset
    )
    df_out["settlement_period"] = (
        df_out["hour"] * 2 + df_out["settlement_offset"] + 1
    )
    df_out = df_out.drop(columns=["settlement_offset"])

    return df_out


def drop_dupes_and_nulls(
    df: pd.DataFrame, drop_nulls: bool = True
) -> pd.DataFrame:
    """
    Function to drop duplicated readings and replace missing readings with 0.0

    Args:
        df (pd.DataFrame): Input dataframe
        drop_nulls (bool): Whether to drop rows with NaN kwh values. If False,
            will replace NaN kwh values with 0.0
    Returns:
        pd.DataFrame: Output dataframe
    """
    logger.info("🗑 Dropping dupes")
    df_out = df.copy()
    df_out = df_out.sort_values(
        by=["ID", "date", "settlement_period"], ascending=True
    )
    df_out = df_out.drop_duplicates(
        subset=["ID", "date", "settlement_period"], keep="last"
    )
    if drop_nulls:
        logger.info("🗑 Dropping nulls")
        df_out = df_out.dropna(subset="kwh")
    else:
        logger.info("🗑 Filling nulls with 0.0")
        df_out["kwh"] = df_out["kwh"].fillna(0.0)
    return df_out


def filter_missing_kwh(
    df: pd.DataFrame, time_resolution: str = "half_hourly"
) -> pd.DataFrame:
    """
    Drop dates where we don't have full 48 readings if half-hourly data,
    or 24 readings if hourly data, for a given ID and date.

    Args:
        df (pd.DataFrame): dataset
        time_resolution (str): Time resolution of the data. Allowed values:
            "half_hourly", "hourly".

    Returns:
        pd.DataFrame: filtered dataset
    """
    logger.info("🔍 Filtering missing kwh readings")
    merge_cols = ["ID", "date"]
    df_group = df.groupby(merge_cols)[["kwh"]].count().reset_index()

    if time_resolution == "half_hourly":
        required_len = 48  # 48 hh readings
    elif time_resolution == "hourly":
        required_len = 24  # 24 h readings
    else:
        raise ValueError(
            f"time_resolution must be 'half_hourly' or 'hourly', \
        got {time_resolution}"
        )
    df_group["required_len"] = required_len

    df_full_data = df_group.query("required_len==kwh")  # Has all required data
    df_out = df_full_data[merge_cols].merge(df, on=merge_cols, how="inner")

    if len(df_out) == 0:
        raise ValueError(
            "No data left after filtering days with missing kWh readings"
        )
    return df_out


def pack_smart_meter_data_into_arrays(
    df: pd.DataFrame, feature_cols: List[str]
) -> pd.DataFrame:
    """
    Pack smart meter data into 48- or 24-dimensional arrays,
    depending on whether half-hourly or hourly data.
    Note: LCL dataset are all given in UTC. We should expect
    48-readings per day.

    Args:
        df (pd.DataFrame): Input dataframe
        feature_cols (List[str]): List of feature columns to include in
            processed dataset.

    Returns:
        pd.DataFrame: Output dataframe
    """
    logger.info("👝 Packing time series into arrays")
    df_out = df.copy()
    df_out = df_out.sort_values(
        by=["ID", "date", "settlement_period"], ascending=True
    )
    groupby_cols = ["ID", "month_end", "month"] + feature_cols
    groupby_cols = groupby_cols + ["dayofweek", "day", "date"]

    df_out = pd.DataFrame(
        df_out.groupby(groupby_cols)["kwh"]
        .agg(lambda x: x.tolist())
        .reset_index()
    )
    return df_out


def get_mean_and_std(df: pd.DataFrame) -> Tuple[float, float]:
    """
    Calculate the mean and standard deviation of the dataset

    Args:
        df (pd.DataFrame): Input dataframe

    Returns:
        Tuple[float, float]: Mean and standard deviation
    """
    logger.info("📊 Calculating mean and std")
    mean = np.mean(df["kwh"])
    std = np.std(df["kwh"])
    return mean, std


def create_outliers(
    df: pd.DataFrame, time_resolution: str, mean: float, mean_factor: int = 20
) -> pd.DataFrame:
    """
    Function to generate outliers based on gaussian and gamma distribution.
    Noise is generated based on a mean of 20 times population mean
    as descibed in the "Defining Good" paper.

    Args:
        df (pd.DataFrame): Input dataframe to sample rows from
        mean (float): Dataset population mean

    Returns:
        pd.DataFrame: Dataframe consisting of noisy outliers
    """

    if time_resolution == "half_hourly":
        n = 48
    elif time_resolution == "hourly":
        n = 24
    else:
        raise ValueError("time_resolution must be 'half_hourly' or 'hourly'")

    gaussian_generator = NoiseFactory(
        noise_type=NoiseType.GAUSSIAN,
        mean=mean,
        scale=1.0,
        mean_factor=mean_factor,
        size=(50, n),
    )

    gamma_generator = NoiseFactory(
        noise_type=NoiseType.GAMMA,
        mean=mean,
        scale=1.0,
        mean_factor=mean_factor,
        size=(50, n),
    )
    logger.info(
        "🎲 Generating unseen outliers with mean:"
        f"{mean:.4f} and mean_factor: {mean_factor:.4f}"
    )
    df_gaussian_noise = gaussian_generator.inject_noise(df)
    df_gamma_noise = gamma_generator.inject_noise(df)

    df_noise = pd.concat([df_gaussian_noise, df_gamma_noise])
    return df_noise


def preprocess_pipeline(
    file_path: Path,
    out_path: Path,
    datetime_col: str = "DateTime",
    kwh_col: str = "kwh",
    id_col: str = "ID",
    utc: bool = True,
    datetime_format: Optional[str] = None,
    time_resolution: str = "half_hourly",
    feature_cols: List[str] = ["stdorToU"],
    drop_nulls: bool = True,
):
    """Preprocess the raw data for Faraday training and evaluation.
    Pipeline includes:
    - Data cleaning
    - Extraction of time features (day of week, month of year)
    - Injection of outliers into dataset
    - Calculation of mean and std for normalisation in Faraday
    - Re-structuring the dataset so that each row corresponds to a daily load
        profile.

    Args:
        file_path (Path): Path to the raw CSV file
        out_path (Path): Path to save the processed data
        datetime_col (str, optional): datetime column. Defaults to "DateTime".
        kwh_col (str, optional): kWh column. Defaults to "kwh".
        id_col (str, optional): Household ID column. Defaults to "ID".
        utc (bool, optional): Whether the datetime is in UTC. Defaults to True.
        datetime_format (str, optional): Format of the datetime column. If
            None, will try to infer the format. Defaults to None.
        time_resolution (str, optional): Time resolution of the data. Allowed
            values: "half_hourly", "hourly". Defaults to "half_hourly".
        feature_cols (List[str], optional): List of feature columns to include.
            Defaults to ["stdorToU"].
        drop_nulls (bool): Whether to drop rows with NaN kwh values. If False,
            will replace NaN kwh values with 0.0
    """

    df = load_data(file_path)
    df = format_data(df, datetime_col, kwh_col, id_col, utc, datetime_format)
    df = extract_date_features(df)
    df = parse_settlement_period(df)
    df = drop_dupes_and_nulls(df, drop_nulls)
    df = filter_missing_kwh(df, time_resolution)

    mean, stdev = get_mean_and_std(df)
    df = pack_smart_meter_data_into_arrays(df, feature_cols)

    df_noise = create_outliers(df, time_resolution, mean)

    os.makedirs(out_path, exist_ok=True)
    df.to_csv(f"{out_path}/data.csv", index=False)
    df_noise.to_csv(f"{out_path}/outliers.csv", index=False)

    mean_std_dict = {"mean": mean, "stdev": stdev}
    with open(f"{out_path}/mean_std.csv", "w") as f:
        w = csv.DictWriter(f, mean_std_dict.keys())
        w.writeheader()
        w.writerow(mean_std_dict)


def preprocess_data(
    data_dir: str,
    datetime_col: str = "DateTime",
    kwh_col: str = "kwh",
    id_col: str = "ID",
    utc: bool = True,
    datetime_format: Optional[str] = None,
    time_resolution: str = "half_hourly",
    feature_cols: List[str] = ["stdorToU"],
    drop_nulls: bool = True,
):

    SOURCE_DIR = f"{data_dir}/raw"
    OUT_DIR = f"{data_dir}/processed"
    preprocess_pipeline(
        file_path=Path(f"{SOURCE_DIR}/historical/train.csv"),
        out_path=Path(f"{OUT_DIR}/historical/train"),
        datetime_col=datetime_col,
        kwh_col=kwh_col,
        id_col=id_col,
        utc=utc,
        datetime_format=datetime_format,
        time_resolution=time_resolution,
        feature_cols=feature_cols,
        drop_nulls=drop_nulls,
    )
    preprocess_pipeline(
        file_path=Path(f"{SOURCE_DIR}/historical/holdout.csv"),
        out_path=Path(f"{OUT_DIR}/historical/holdout"),
        datetime_col=datetime_col,
        kwh_col=kwh_col,
        id_col=id_col,
        utc=utc,
        datetime_format=datetime_format,
        time_resolution=time_resolution,
        feature_cols=feature_cols,
        drop_nulls=drop_nulls,
    )
    preprocess_pipeline(
        file_path=Path(f"{SOURCE_DIR}/future/train.csv"),
        out_path=Path(f"{OUT_DIR}/future/train"),
        datetime_col=datetime_col,
        kwh_col=kwh_col,
        id_col=id_col,
        utc=utc,
        datetime_format=datetime_format,
        time_resolution=time_resolution,
        feature_cols=feature_cols,
        drop_nulls=drop_nulls,
    )
    preprocess_pipeline(
        file_path=Path(f"{SOURCE_DIR}/future/holdout.csv"),
        out_path=Path(f"{OUT_DIR}/future/holdout"),
        datetime_col=datetime_col,
        kwh_col=kwh_col,
        id_col=id_col,
        utc=utc,
        datetime_format=datetime_format,
        time_resolution=time_resolution,
        feature_cols=feature_cols,
        drop_nulls=drop_nulls,
    )



# Copyright Contributors to the Opensynth-energy Project.
# SPDX-License-Identifier: Apache-2.0

from typing import Tuple

import torch
import torch.nn.functional as F


def _expand_samples(x: torch.Tensor, weights: torch.Tensor) -> torch.Tensor:
    """
    Expand the repeat in the tensor based on the sample weights
    e.g. if x: [1,2,3] and weights: [1,2,3], the output will be [1,2,2,3,3,3]

    Args:
        x (torch.Tensor): Input tensor
        weights (torch.Tensor): Sample weight

    Returns:
        torch.Tensor: Repeated tensor by sample weight
    """
    weights = weights.squeeze().long()
    return torch.repeat_interleave(x, weights, dim=0)


def _check_shape(x: torch.Tensor) -> torch.Tensor:
    """
    Check that tensor is either a 1-D or 2_D tensor.
    If tensor is 1-D, reshape to [N,1] tensor.

    Args:
        x (torch.Tensor): Input tensor

    Raises:
        ValueError: Raises error if tensor is not 1-D or 2-D

    Returns:
        torch.Tensor: Reshaped tensor
    """
    if len(x.shape) == 1:
        return x.reshape(-1, 1)
    elif len(x.shape) == 2:
        return x
    else:
        raise ValueError(f"Input tensor {x} must be 1D or 2D")


def mmd_loss(
    y: torch.Tensor, x: torch.Tensor, sample_weights: torch.Tensor = None
) -> torch.Tensor:
    """
    Calculate MMD Loss

    Args:
        y (torch.Tensor): Tensor Y
        x (torch.Tensor): Tensor X

    Returns:
        torch.Tensor: MMD Distances between Tensor Y and X
    """
    # Expand the weights array based on number of samples
    xx, yy, zz = torch.mm(x, x.t()), torch.mm(y, y.t()), torch.mm(x, y.t())
    rx = xx.diag().unsqueeze(0).expand_as(xx)
    ry = yy.diag().unsqueeze(0).expand_as(yy)

    dxx = rx.t() + rx - 2.0 * xx
    dyy = ry.t() + ry - 2.0 * yy
    dxy = rx.t() + ry - 2.0 * zz

    XX, YY, XY = (
        torch.zeros(xx.shape),
        torch.zeros(xx.shape),
        torch.zeros(xx.shape),
    )
    XX = XX.to(x)
    YY = YY.to(x)
    XY = XY.to(x)

    bandwidth_range = [10, 15, 20, 50]
    for a in bandwidth_range:
        XX += torch.exp(-0.5 * dxx / a)
        YY += torch.exp(-0.5 * dyy / a)
        XY += torch.exp(-0.5 * dxy / a)

    mmd_loss = XX + YY - 2.0 * XY

    if sample_weights is not None:
        weights = sample_weights.squeeze().long()
        mmd_loss = torch.repeat_interleave(mmd_loss, weights, dim=0)
        mmd_loss = mmd_loss * weights
        return torch.sum(mmd_loss) / torch.sum(sample_weights) ** 2

    return torch.mean(mmd_loss)


def quantile_loss(
    y_pred: torch.Tensor,
    y_real: torch.Tensor,
    quantile: float,
    sample_weights: torch.Tensor = None,
) -> torch.Tensor:
    """
    Calculate quantile loss
    Args:
        y_pred (torch.Tensor): Predicted quantile
        y_real (torch.Tensor): Actual quantile
        quantile (float): Quantile value

    Returns:
        torch.Tensor: Quantile loss
    """
    y_pred = _check_shape(y_pred)
    y_real = _check_shape(y_real)
    quantile_loss = torch.max(
        quantile * (y_real - y_pred), (1 - quantile) * (y_pred - y_real)
    )

    if sample_weights is not None:
        quantile_loss = quantile_loss * sample_weights.reshape(
            len(sample_weights), 1
        )

        return torch.sum(quantile_loss) / (
            torch.sum(sample_weights) * y_pred.shape[1]
        )

    return torch.mean(quantile_loss)


def mse_loss(
    y_pred: torch.Tensor,
    y_real: torch.Tensor,
    sample_weights: torch.Tensor = None,
):
    """
    Calculate MSE loss. If sample weights are provided, sample losses are
    multiplied by the sample weight before averaging.

    Args:
        y_pred (torch.Tensor): Prediction
        y_real (torch.Tensor): Actual
        sample_weights (torch.Tensor, optional): Sample Weights.
        Defaults to None.

    Returns:
        _type_: _description_
    """
    y_pred = _check_shape(y_pred)
    y_real = _check_shape(y_real)
    squared_loss = F.mse_loss(y_pred, y_real, reduction="none")

    if sample_weights is not None:
        squared_loss = squared_loss * sample_weights.reshape(
            len(sample_weights), 1
        )
        return torch.sum(squared_loss) / (
            torch.sum(sample_weights) * y_pred.shape[1]
        )

    return torch.mean(squared_loss)


def calculate_training_loss(
    x_hat: torch.Tensor,
    x: torch.Tensor,
    mse_weight: float,
    quantile_upper_weight: float,
    quantile_lower_weight: float,
    quantile_median_weight: float,
    lower_quantile: float,
    upper_quantile: float,
    sample_weights: torch.Tensor = None,
) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor]:
    """
    Calculate training losses for Faraday.
    Losses are a combined total of:
    1) MMD loss
    2) MSE loss * mse_weight
    3) upper quantile loss * upper quantile weight
    4) lower quantile loss * lower quantile weight
    5) median quantile loss * median quantile weight
    Args:
        x_hat (torch.Tensor): Generated tensor
        x (torch.Tensor): Real tensor
        mse_weight (float): MSE Weight
        quantile_upper_weight (float): Upper quantile weight
        quantile_lower_weight (float): Lower quantile weight
        quantile_median_weight (float): Median quantile weight
        lower_quantile (float): Lower quantile value
        upper_quantile (float): Upper quantile value

    Returns:
        Tuple[torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor]:
        Total loss, MMD loss, MSE loss, Quantile loss
    """
    mmd_loss_ = mmd_loss(x_hat, x, sample_weights=sample_weights)
    mse_loss_ = mse_loss(x_hat, x, sample_weights=sample_weights) * mse_weight
    quantile_upper_loss = (
        quantile_loss(x_hat, x, upper_quantile, sample_weights)
        * quantile_upper_weight
    )
    quantile_lower_loss = (
        quantile_loss(x_hat, x, lower_quantile, sample_weights)
        * quantile_lower_weight
    )
    quantile_median_loss = (
        quantile_loss(x_hat, x, 0.5, sample_weights) * quantile_median_weight
    )
    quantile_losses = (
        quantile_upper_loss + quantile_lower_loss + quantile_median_loss
    )

    total_loss = mmd_loss_ + mse_loss_ + quantile_losses

    return total_loss, mmd_loss_, mse_loss_, quantile_losses



# Copyright Contributors to the Opensynth-energy Project.
# SPDX-License-Identifier: Apache-2.0

import logging
from typing import Optional, Union

import numpy as np
import numpy.typing as npt
import pytorch_lightning as pl
import torch

from opensynth.data_modules.lcl_data_module import LCLDataModule, TrainingData
from opensynth.data_modules.streaming_data_module import StreamDataModule
from opensynth.models.faraday.gaussian_mixture import (
    GaussianMixtureLightningModule,
    GaussianMixtureModel,
    initialise_gmm_params,
)
from opensynth.models.faraday.vae_model import FaradayVAE

logger = logging.getLogger(__name__)


class FaradayModel:

    def __init__(
        self,
        vae_module: FaradayVAE,
        n_components: int,
        tol: float = 1e-3,
        accelerator: str = "cpu",
        devices: int = 1,
        max_epochs: int = 1000,
        covariance_reg: float = 1e-6,
        sample_weights_column: Optional[str] = None,
    ):
        """
        Faraday Model. Note:

        - Faraday Model only supports integer-encoded labels.
        - If you wish to have float labels, you should subclass FaradayModel
        and implement your own `sample_gmm` method.
        - Faraday Model also expects data module to return a
        TrainingData object.

        Args:
            vae_module (FaradayVAE): Trained VAE component.
            n_components (int): GMM clusters.
            tol (float, optional): Tolerance for GMM. Defaults to 1e-3.
            accelerator (str, optional): Accelerator for GMM training.
                Defaults to "cpu".
            devices (int, optional): Number of devices for GMM training.
                Defaults to 1.
            max_epochs (int, optional): Max epochs for GMM training.
                Defaults to 1000.
            covariance_reg (float, optional): Covariance
                regularization for GMM. Defaults to 1e-6.
                This is added to the diagonal of the covariance matrix to
                ensure that it is positive semi-definite. Higher values will
                make the algorithm more robust to singular covariance matrices,
                at the cost of higher regularization.
        """
        self.n_components = n_components
        self.vae_module = vae_module
        self.tol = tol
        self.accelerator = accelerator
        self.devices = devices
        self.max_epochs = max_epochs
        self.covariance_reg = covariance_reg
        self.sample_weights_column = sample_weights_column

    @staticmethod
    def parse_samples(
        samples: torch.Tensor, latent_dim: int, feature_list: list[str]
    ) -> TrainingData:
        """
        Abstract the labels and round numerical values to integers
        Order of features needs to be preserved so VAE can decode
        it correctly, and is handled by `get_index` method.

        Args:
            samples (torch.Tensor): Samples sampled from GMM
            latent_dim (int): Latent Dimension Size
            feature_list (list[str]): List of feature names

        Returns:
            TrainingData: TypedDict containing GMM sampled
            kWh and feature labels
        """

        kwh: torch.Tensor = samples[:, :latent_dim]
        labels: dict[str, torch.Tensor] = {}

        for i, feature in enumerate(feature_list):
            index = FaradayModel.get_index(feature_list, i)
            feature_tensor = torch.round(
                torch.Tensor(samples[:, index]), decimals=0
            ).int()
            labels[feature] = feature_tensor.reshape(len(kwh), 1)

        return TrainingData(kwh=kwh, features=labels)

    @staticmethod
    def get_feature_range(
        features: dict[str, torch.Tensor]
    ) -> dict[str, dict[str, int]]:
        """
        Get the max and min values of numerically encoded features
        Args:
            features (dict[str, torch.Tensor]): Dictionary of
            feature tensors
        Returns:
            dict[str, dict[str, int]]: A dictionary of
            min and max values for each feature label
        """
        feature_range: dict[str, dict[str, int]] = {}
        for feature in features:
            feature_range[feature] = {
                "min": features[feature].min().item(),
                "max": features[feature].max().item(),
            }
        return feature_range

    @staticmethod
    def create_mask(
        gmm_labels: dict[str, torch.Tensor],
        range_dict: dict[str, dict[str, int]],
    ) -> npt.NDArray[np.bool_]:
        """
        Create a filter mask to make sure that GMM sampled labels are within
        the range of accepted values. There is no guarantee that GMM will not
        create out-of-distribution values. For example if `month of year`
        ranges from 0-11, there is no guarantee that GMM will not
        result in `month of year` = 100.
        Args:
            gmm_labels (dict[str, torch.Tensor]): Dictionary of features
            obtained from GMM model
            range_dict (dict[str, dict[str, int]]): Dictionary of
            ranges of each label
        Returns:
            list[bool]: Mask to filter out-of-distribution values
        """
        label_mask: Optional[npt.NDArray[np.bool_]] = None  # Initialize mask

        for feature, bounds in range_dict.items():

            # Fetch the feature range
            min_value = bounds.get("min")
            max_value = bounds.get("max")
            if min_value is None or max_value is None:
                raise ValueError(f"Feature {feature} have missing range")

            # Dynamically fetch the respective labels
            gmm_value: torch.Tensor = gmm_labels.get(feature)
            if gmm_value is None:
                raise ValueError(f"Feature {feature} not found in GMM labels")

            # Create the mask for this label
            current_mask = np.logical_and(
                gmm_value.detach().numpy() >= min_value,
                gmm_value.detach().numpy() <= max_value,
            )

            # Combine the masks with `&` (AND operation)
            label_mask = (
                current_mask
                if label_mask is None
                else (label_mask & current_mask)
            )

        label_mask = np.squeeze(label_mask)
        return np.squeeze(label_mask)

    @staticmethod
    def get_index(feature_list: list[str], current_index: int) -> int:
        """
        Get the index of each label, so that we can store the
        correct column as the correct label in the gmm_labels
        dictionary which the VAE needs in order to decode.

        Args:
            feature_list (list[str]): List of features
            current_index (int): Current iter index

        Returns:
            int: Returns colum index
        """
        return -(len(feature_list) - current_index)

    @staticmethod
    def filter_mask(
        mask: npt.NDArray[np.bool_], sampled_data: TrainingData
    ) -> TrainingData:
        """
        Given a mask and GMM sampled data, filter the GMM
        sample with the mask.

        Args:
            mask (npt.NDArray[np.bool_]): Numpy array of boolean values
            sampled_data (TrainingData): GMM sampled data

        Returns:
            TrainingData: Filtered GMM samples
        """
        sampled_data["kwh"] = sampled_data["kwh"][mask]
        for feature in sampled_data["features"]:
            sampled_data["features"][feature] = sampled_data["features"][
                feature
            ][mask]
        return sampled_data

    def train_gmm(self, dm: Union[StreamDataModule, LCLDataModule]):
        """
        Train Gaussian Mixture Module

        Args:
            dm (StreamDataModule, LCLDataModule): Training data
        """
        logger.info("🚀 Training GMM")

        dl = dm.train_dataloader()
        next_batch = next(iter(dl))

        # Explicit check to make sure that data module used
        # to train GMM has features ordered the same way as
        # when training the VAE
        features = next_batch["features"]
        obtained_feature_list = list(features.keys())
        num_features = self.vae_module.latent_dim + len(obtained_feature_list)

        expected_feature_list = self.vae_module.feature_list
        if obtained_feature_list != expected_feature_list:
            logger.error(
                """Feature list required by `vae_module` and does not match
                features specified by the data module.
                """
            )

        # Initialise GMM parameters
        # Initialising on the first batch of the data only
        # If data is shuffled, this should represent the full data distribution
        gmm_init_params = initialise_gmm_params(
            next_batch,
            n_components=self.n_components,
            vae_module=self.vae_module,
            reg_covar=self.covariance_reg,
            sample_weights_column=self.sample_weights_column,
        )

        gmm_module = GaussianMixtureModel(
            num_components=self.n_components,
            num_features=num_features,
            reg_covar=self.covariance_reg,
        )
        gmm_module.initialise(gmm_init_params)
        print(
            f"Initial prec chol: {gmm_module.precision_cholesky[0][0][0]}. \
                Initial mean: {gmm_module.means[0][0]}"
        )

        # Fit GMM
        gmm_lightning_module = GaussianMixtureLightningModule(
            gmm_module=gmm_module,
            vae_module=self.vae_module,
            num_components=gmm_module.num_components,
            num_features=gmm_module.num_features,
            reg_covar=gmm_module.reg_covar,
            convergence_tolerance=self.tol,
            sync_on_batch=False,
            sample_weights_column=self.sample_weights_column,
        )
        trainer = pl.Trainer(
            max_epochs=self.max_epochs,
            accelerator="cpu",
            deterministic=True,
        )
        trainer.fit(gmm_lightning_module, dl)

        # Record the ranges of features seen during training
        # Assuming that batches are random, than this should
        # be representative.
        self.feature_range = self.get_feature_range(features)
        logger.info("🎉 GMM Training Completed")

        self.gmm_module = gmm_module

    def sample_gmm(self, n_samples: int) -> TrainingData:
        """
        Samples latent codes from GMM and decode with decoder.

        Args:
            n_samples (int): Number of samples to generate.

        Returns:
            TrainingData: Decoder output (KWH), feature labels
        """

        gmm_samples: torch.Tensor = self.gmm_module.sample(n_samples)

        # Parse GMM samples
        gmm_samples_parsed: TrainingData = self.parse_samples(
            gmm_samples,
            self.vae_module.latent_dim,
            self.vae_module.feature_list,
        )

        # Filter invalid (out of distribution) samples
        label_mask = self.create_mask(
            gmm_samples_parsed["features"], self.feature_range
        )
        gmm_samples_filtered = self.filter_mask(label_mask, gmm_samples_parsed)

        # Decode samples with VAE
        decoder_input = self.vae_module.reshape_data(
            gmm_samples_filtered["kwh"], gmm_samples_filtered["features"]
        ).float()

        decoder_output = self.vae_module.decode(decoder_input)

        # Output samples
        outputs = TrainingData(
            kwh=decoder_output, features=gmm_samples_filtered["features"]
        )
        return outputs



# Copyright Contributors to the Opensynth-energy Project.
# SPDX-License-Identifier: Apache-2.0
#
# The EnergyDiff model is made available via Nan Lin and Pedro P. Vergara
# from the Delft University of Technology. Nan Lin and Pedro P. Vergara are
# funded via the ALIGN4Energy Project (with project number NWA.1389.20.251) of
# the research programme NWA ORC 2020 which is (partly) financed by the Dutch
# Research Council (NWO), The Netherland.
"""
Name: diffusion base class and utility functions
Author: Nan Lin (sentient-codebot)
Date: Nov 2024

"""
import enum
import math
from abc import ABC, abstractmethod
from collections import namedtuple
from typing import Iterator

import torch
from torch import Tensor, nn
from torch.nn.parameter import Parameter

ModelPrediction = namedtuple(
    "ModelPrediction", ["pred_noise", "pred_x_start", "pred_var_factor"]
)


class ModelMeanType(enum.Enum):
    X_START = "x_start"
    NOISE = "noise"
    V = "v"


class ModelVarianceType(enum.Enum):
    FIXED_SMALL = "fixed_small"  # default, necessary for dpm-solver
    FIXED_LARGE = "fixed_large"
    LEARNED_RANGE = "learned_range"


class LossType(enum.Enum):
    MSE = "mse"
    RESCALED_MSE = "rescaled_mse"
    KL = "kl"
    RESCALED_KL = "rescaled_kl"

    def is_vb(self):
        return self in (LossType.KL, LossType.RESCALED_KL)


class BetaScheduleType(enum.Enum):
    DISCRETE = "discrete"
    LINEAR = "linear"
    COSINE = "cosine"  # default


def extract(a: Tensor, t: Tensor, x_shape: torch.Size) -> Tensor:
    """Extract values from tensor 'a' at positions specified by tensor 't' \
        and reshape to 'x_shape'.

    This function extracts values from tensor 'a' at positions specified \
        by tensor 't',
    and reshapes the result to match the shape specified by 'x_shape'.

    Args:
        a (Tensor): Source tensor to extract values from. 1D tensor.
        t (Tensor): Tensor containing positions where values should be \
            extracted. (batch,) or (batch, 1) tensor.
        x_shape (torch.Size): Target shape for the output tensor.

    Returns:
        Tensor: Extracted values reshaped to match x_shape.

    Example:
        >>> a = torch.tensor([0.1, 0.2, 0.3, 0.4])
        >>> t = torch.tensor([1, 3])  # extract values at indices 1 and 3
        >>> x_shape = (2, 3, 2)  # target shape
        >>> result = extract(a, t, x_shape)
        >>> result.shape
        torch.Size([2, 3, 2])
        >>> result[0, :, :]  # filled with 0.2 (a[1])
        tensor([[0.2, 0.2],
                [0.2, 0.2],
                [0.2, 0.2]])
        >>> result[1, :, :]  # filled with 0.4 (a[3])
        tensor([[0.4, 0.4],
                [0.4, 0.4],
                [0.4, 0.4]])

    Note:
        The function automatically converts tensor 'a' to float32 dtype and
        matches the device of tensor 't'.
    """
    # shape chec
    if t.dim() > 2 or (t.dim() == 2 and t.shape[1] != 1):
        raise ValueError("t must be a 1D tensor or 2D tensor with shape (B,1)")

    if t.dim() == 2:
        t = t.squeeze(-1)  # Convert 2D (B,1) to 1D (B,)

    if t.shape[0] != x_shape[0]:
        raise ValueError("t and x_shape must have the same batch size")

    # convert a to float32 and match device
    a = a.to(device=t.device, dtype=torch.float32)  # !dtype is fixed here
    b, *_ = t.shape
    dim_x = len(x_shape)
    # extract values from a at positions specified by t
    out = a.gather(-1, t)
    # fill missing dimensions with singleton dimensions (braodcast)
    target_shape = (b, *(1 for _ in range(dim_x - 1)))
    # using reshape to covert to (b, 1, ..., 1)
    # and use "+" to broadcast to target shape
    out = out.reshape(*target_shape) + torch.zeros(x_shape, device=t.device)

    return out


def linear_beta_schedule(num_timestep: int) -> Tensor:
    """get the linear beta schedule.

    Arguments:
        num_timestep -- int

    Returns:
        Tensor -- shape: (num_timestep,)
    """
    scale = 1000.0 / num_timestep
    beta_start = scale * 1e-4
    beta_end = scale * 2e-2
    return torch.linspace(
        beta_start, beta_end, num_timestep, dtype=torch.float64
    )  # shape: (num_timestep,)


def cosine_beta_schedule(num_timestep: int, s: float = 0.008) -> Tensor:
    "as proposed in https://openreview.net/forum?id=-NEXDKk8gZ"
    num_step = num_timestep + 1
    x = torch.linspace(
        0, num_timestep, num_step, dtype=torch.float64
    )  # shape: (num_step,)
    alpha_cumprod = (
        torch.cos((x / num_timestep + s) / (1 + s) * math.pi / 2) ** 2
    )  # shape: (num_step,)
    alpha_cumprod = alpha_cumprod / alpha_cumprod[0]  # shouldn't [0] be 1.?
    beta_schedule = 1 - (
        alpha_cumprod[1:] / alpha_cumprod[:-1]
    )  # shape: (num_timestep,)
    return torch.clip(beta_schedule, 0, 0.999)  # shape: (num_timestep,)


def get_beta_schedule(
    beta_schedule_type: BetaScheduleType, num_timestep: int
) -> Tensor:
    if beta_schedule_type == BetaScheduleType.LINEAR:
        return linear_beta_schedule(num_timestep)
    elif beta_schedule_type == BetaScheduleType.COSINE:
        return cosine_beta_schedule(num_timestep)
    else:
        raise ValueError("type_beta_schedule must be one of linear, cosine")


def get_loss_weight(
    alpha_cumprod: Tensor, model_mean_type: ModelMeanType
) -> Tensor:
    # TODO unit test
    # calculate loss weights
    snr = alpha_cumprod / (1.0 - alpha_cumprod)
    if model_mean_type == ModelMeanType.NOISE:
        loss_weight = torch.ones_like(snr)
    elif model_mean_type == ModelMeanType.X_START:
        loss_weight = snr
    elif model_mean_type == ModelMeanType.V:
        loss_weight = snr / (snr + 1.0)

    return loss_weight


class DiffusionBase(ABC):
    @property
    @abstractmethod
    def alpha_cumprod(self) -> Tensor:
        pass

    @property
    @abstractmethod
    def model_mean_type(self) -> ModelMeanType:
        pass

    @property
    @abstractmethod
    def model_variance_type(self) -> ModelVarianceType:
        pass

    @property
    @abstractmethod
    def model(self) -> nn.Module:
        pass

    @abstractmethod
    def parameters(self) -> Iterator[Parameter]:
        pass



# Copyright Contributors to the Opensynth-energy Project.
# SPDX-License-Identifier: Apache-2.0

from typing import Optional

import torch

from opensynth.data_modules.lcl_data_module import TrainingData
from opensynth.models.faraday.gaussian_mixture import gmm_model, gmm_utils
from opensynth.models.faraday.vae_model import FaradayVAE


def initialise_gmm_params(
    data: TrainingData,
    n_components: int,
    vae_module: FaradayVAE,
    reg_covar: float = 1e-6,
    sample_weights_column: Optional[str] = None,
) -> gmm_model.GMMInitParams:
    """
    Initialise Gaussian Mixture Parameters.
    Component means and component assignments are initialised using K-means,
      and porting SKLearn's implementation of computing cholesky precision and
      covariances.

    Args:
        data (TrainingData): Input data
        n_components (int): Number of components
        vae_module(FaradayVAE): Trained VAE model
        reg_covar (float): Regularisation for covariance matrix

    Returns:
        dict[str, torch.Tensor]: GMM params
    """

    encoded_data = gmm_utils.prepare_data_for_training_step(
        data, vae_module, sample_weights_column
    )
    X = encoded_data.detach().numpy()

    labels_, means_, responsibilities_ = gmm_utils.initialise_centroids(
        X=X, n_components=n_components
    )

    weights_, means_, covariances_ = (
        gmm_utils.torch_estimate_gaussian_parameters(
            X=torch.from_numpy(X),
            responsibilities=responsibilities_,
            reg_covar=reg_covar,
        )
    )

    precision_cholesky_ = gmm_utils.torch_compute_precision_cholesky(
        covariances=covariances_, reg=reg_covar
    )

    init_params = gmm_model.GMMInitParams(
        init_labels=labels_,
        means=means_,
        weights=weights_,
        covariances=covariances_,
        precision_cholesky=precision_cholesky_,
    )
    return init_params
